--[[
Name: LibSpreadsheet-1.0
Revision: $Rev: 51 $
Author(s): Humbedooh
Description: Spreadsheet stuff
Dependencies: LibStub AceGUI-3.0
License: GPL v3 or later.
]]
-----------------------------------------------------------------------
if ( dofile ) then
    dofile([[..\LibStub\LibStub.lua]]);
end
local SPREADSHEET = "LibSpreadsheet-1.0"
local SPREADSHEET_MINOR = tonumber(("$Rev: 051 $"):match("(%d+)")) or 9999;
if not LibStub then error(SPREADSHEET .. " requires LibStub.") end
LibSpreadsheet = LibStub:NewLibrary(SPREADSHEET, SPREADSHEET_MINOR)
if not LibSpreadsheet then return end
--[[
do
    if ( debugstack ) then
        local path = string.match(debugstack(1,1,0), "AddOns\\(.+)LibSpreadsheet%-1%.0%.lua")
        if path then
            LibSpreadsheet.directory = "Interface\\AddOns\\"..path
        else
            error(SPREADSHEET.." cannot determine the folder it is located in because the path is too long and got truncated in the debugstack(1,1,0) function call")
        end
    end
end
]]--
if ( table.wipe == nil ) then
	table.wipe = function(obj) obj = {}; end;
end
--[[
LibSpreadsheet-1.0 DEFAULTS
]]--
LibSpreadsheet.disabled = LibSpreadsheet.disabled or nil;
LibSpreadsheet.objects = LibSpreadsheet.objects or {};
LibSpreadsheet.cols = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z', '?'};
LibSpreadsheet.frameBuffer = {};
LibSpreadsheet.sortAscending = 1;
LibSpreadsheet.sortDescending = 2;
LibSpreadsheet.sortAscendingWithCase = 3;
LibSpreadsheet.sortDescendingWithCase = 4;
LibSpreadsheet.callbackEvents = { 'OnSelectionChange', 'OnUserInputProcessed', 'OnUserInputRequested', 'OnUserSortRequested', 'OnUserSortProcessed' }
LibSpreadsheet.macroFuncs = {
    {func = "", internal = false, caller = ""},
    {func = "NUM", internal = false, caller = "tonumber"},
    {func = "LEN", internal = false, caller = "string.len"},
    {func = "MAX", internal = false, caller = "math.max"},
    {func = "MIN", internal = false, caller = "math.min"},
    {func = "ABS", internal = false, caller = "math.abs"},--
    {func = "SIN", internal = false, caller = "math.sin"},
    {func = "SINH", internal = false, caller = "math.sinh"},
    {func = "TAN", internal = false, caller = "math.tan"},
    {func = "TANH", internal = false, caller = "math.tanh"},
    {func = "COS", internal = false, caller = "math.cos"},--
    {func = "COSH", internal = false, caller = "math.cosh"},--
    {func = "LN", internal = false, caller = "math.log"},
    {func = "LOG", internal = false, caller = "_G.LibSpreadsheet:Macro_Logarithm"},
    {func = "POW", internal = false, caller = "math.pow"},
    {func = "SQRT", internal = false, caller = "math.sqrt"},
    {func = "MOD", internal = false, caller = "math.fmod"},
    {func = "FLOOR", internal = false, caller = "_G.LibSpreadsheet:Macro_Floor"},--
    {func = "CEIL", internal = false, caller = "_G.LibSpreadsheet:Macro_Ceil"},--
    {func = "SUM", internal = false, caller = "_G.LibSpreadsheet:Macro_Sum"},
    {func = "STDEV", internal = false, caller = "_G.LibSpreadsheet:Macro_StdDev"},
    {func = "LIST", internal = true, caller = "_G.LibSpreadsheetMacro:ListCells"},
    {func = "UNIQUE", internal = false, caller = "_G.LibSpreadsheet:Macro_Unique"},
    {func = "IF", internal = false, caller = "_G.LibSpreadsheet:Macro_If"},
    {func = "AND", internal = false, caller = "_G.LibSpreadsheet:Macro_And"},
    {func = "OR", internal = false, caller = "_G.LibSpreadsheet:Macro_Or"},
    {func = "LARGE", internal = false, caller = "_G.LibSpreadsheet:Macro_Large"},
    {func = "SMALL", internal = false, caller = "_G.LibSpreadsheet:Macro_Small"},
    {func = "MEDIAN", internal = false, caller = "_G.LibSpreadsheet:Macro_Median"},
    {func = "AVERAGE", internal = false, caller = "_G.LibSpreadsheet:Macro_Average"},
    {func = "MEAN", internal = false, caller = "_G.LibSpreadsheet:Macro_Average"},
    {func = "MODE", internal = false, caller = "_G.LibSpreadsheet:Macro_Mode"},
    {func = "ROUND", internal = false, caller = "_G.LibSpreadsheet:Macro_Round"}

    };

local function CreateCellFrame(t, n, o, s)
    local frame = nil;
    if ( #LibSpreadsheet.frameBuffer > 0 ) then
        frame = LibSpreadsheet.frameBuffer[#LibSpreadsheet.frameBuffer];
        table.remove(LibSpreadsheet.frameBuffer);
        if ( o ) then frame:SetParent(o); end
        frame:Show();
        frame.icon:SetTexture(nil);
        frame.icon:Hide();
    else
        frame = _G.CreateFrame(t,n,o,s);
        frame.label = frame:CreateFontString(nil, "DIALOG");
        frame.icon = frame:CreateTexture(nil, "DIALOG");
        frame.background = frame:CreateTexture(nil, "BACKGROUND");
        frame.overlay = frame:CreateTexture(nil, "OVERLAY");
    end
    return frame;
end

local function DeleteCellFrame(frame)
    frame:SetParent(nil);
    frame:Hide();
    frame:SetScript("OnMouseDown", nil);
    frame:SetScript("OnMouseUp", nil);
    frame:SetScript("OnEnter", nil);
    frame:SetScript("OnLeave", nil);
    frame:SetScript("OnDoubleclick", nil);
    table.insert(LibSpreadsheet.frameBuffer, frame);
end

function LibSpreadsheet:RunMacro(macro)
    return self:ParseMacro(macro, true);
end
function LibSpreadsheet:ParseMacro(macro, execute)
    local result = nil;
    local i = 0;
    local n = 1;
    local c;
    local f = "";
    local fc = "";
    local args = "";
    local chunks = {};
    local from = nil;
    while i <= macro:len() do
        i = i + 1;
        c = macro:sub(i,i):upper();
        if ( c == ":" ) then
            if ( f:len() > 0 ) then
                from = f .. fc;
                f = "";
                fc = "";
                args = "";
            end
        end
        if ( c == " " or c == "\t" ) then
            if ( f:len() > 0 and from ~= nil ) then
                table.insert(chunks, {type="func", val='LIST', args={from, f..fc}});
                from = nil;
                f = "";
                fc = "";
                args = "";
            elseif ( f:len() > 0 ) then
                if ( fc ~= "" ) then
                    table.insert(chunks, {type="cell", val=f, args=fc});
                else
                    table.insert(chunks, {type="func", val='LIST', args={f}});
                end
                f = "";
                fc = "";
                args = "";
            elseif ( fc:len() > 0 ) then
                table.insert(chunks, {type="func", val='NUM', args={fc}});
                f = "";
                fc = "";
                args = "";
            end
        elseif ( c == "+" or c == "-" or c == "/" or c == "*" or c == "=" ) then
            if ( f:len() > 0 ) then
                table.insert(chunks, {type="cell", val=f, args=fc});
                f = "";
                fc = "";
                args = "";
            elseif ( fc:len() > 0 ) then
                table.insert(chunks, {type="func", val='NUM', args={fc}});
                f = "";
                fc = "";
                args = "";
            end
            table.insert(chunks, {type="control", val=c, args=nil});
        elseif ( c >= "A" and c <= "Z" ) then f = f .. c;
        elseif ( c == "(" ) then
            local p = 0;
            local si = i;
            local spos = 2;
            local epos = 0;
            local argList = {};
            while si <= macro:len() do
                epos = epos + 1;
                local c = macro:sub(si,si);
                args = args .. c;
                if ( c == "(" ) then p = p + 1; end
                if ( c == ")" ) then p = p - 1; end
                if ( c == "," and p == 1 ) then
                    table.insert(argList, args:sub(spos, epos-1));
                    spos = epos + 1;
                end
                if ( p == 0 ) then
                    break;
                end
                si = si + 1;
            end
            if ( p ~= 0 ) then error("Expected ')' before <eol> in Macro");
            else
     --           print("Got args: " .. args .. ", ended at " .. si);
                i = si
                if ( spos ~= args:len() ) then
                    table.insert(argList, args:sub(spos, epos-1));
                end
                table.insert(chunks, {type="func", val=f, args=argList});
                f = "";
                fc = "";
                args = "";
            end
        elseif ( c >= "0" and c <= "9" or c == ".") then
            fc = fc .. c;
        elseif ( c == "$" ) then
        else
    --        print("Bad char " .. c);
        end
    end
    if ( f:len() > 0 and from ~= nil ) then
        table.insert(chunks, {type="func", val='LIST', args={from, f..fc}});
        from = nil;
        f = "";
        fc = "";
        args = "";
    elseif ( f:len() > 0) then
        if ( fc ~= "" ) then
                    table.insert(chunks, {type="cell", val=f, args=fc});
                else
                    table.insert(chunks, {type="func", val='LIST', args={f}});
                end
        f = "";
        fc = "";
        args = "";
    elseif ( fc:len() > 0 ) then
        table.insert(chunks, {type="func", val='NUM', args={fc}});
        f = "";
        fc = "";
        args = "";
    end
    local fstring = "";
    for k, func in pairs(chunks) do
        if ( func.type == "control" ) then
            if ( func.val == "=" ) then func.val = "=="; end
                fstring = fstring .. " " .. func.val .. " ";
        end
        if ( func.type == "cell" ) then
            fstring = fstring .. "_G.LibSpreadsheetMacro:GetValue('" .. func.val .. "', " .. func.args .. ") ";
        end
        if ( func.type == "func" ) then
            local argsCompiled = {};
            local realFunc = {};
            for k, entry in pairs (LibSpreadsheet.macroFuncs) do
                if ( func.val == entry.func ) then
                    realFunc = entry;
                    break;
                end
            end
            if ( realFunc.internal == nil ) then
                error("No such macro function: " .. func.val .. "!");
            end
            if ( realFunc.internal == false ) then
                for k, arg in pairs(func.args) do
                    if ( arg:len() > 1 and tonumber(arg) == nil ) then
         --               print("compiling argument " .. arg);
                        table.insert(argsCompiled, self:ParseMacro(arg, false));
                    elseif ( arg:len() == 1 and arg >= "A" and arg <= "Z" ) then
                        table.insert(argsCompiled, "_G.LibSpreadsheetMacro:ListCells('" .. arg .. "')");
                    else
                        table.insert(argsCompiled, arg);
                    end
                end
            else
                for k, arg in pairs(func.args) do
                    if ( arg:len() > 1 and tonumber(arg) ~= nil ) then
                        table.insert(argsCompiled, self:ParseMacro(arg, false));
                    else
                        table.insert(argsCompiled, '"' .. arg .. '"');
                    end
                end
            end
            fstring = fstring .. realFunc.caller .. "(" .. table.concat(argsCompiled, ", ") .. ")";
        end
        --print("Func: " .. func.val.. ", args: " .. (func.args or "nil"));
    end
  --  print("compiled string: " .. fstring);
    _G.LibSpreadsheetMacro = self;
    if ( not execute ) then
        return fstring;
    else
        result, err = loadstring("return (" ..fstring .. ") ");
        if ( err ) then
            error(err);
   --         print("Error: " .. err);
        else
  --          print("Result: " .. (result() or 0));
            return result();
        end
    end
end


function LibSpreadsheet:Export(format, from, to)
    local rowTable = self.rows;
    if ( self.hasFilters or self.hasGroups ) then
        rowTable = self.rowHeap;
    end
    if ( from == true ) then
        if ( self.selectStartRow ) then
            from = self.selectStartRow;
            to = self.selectEndRow;
        else
            from = 1;
            to = #rowTable;
        end
    end
    local tbl = {};
    for i = math.min(from, to), math.max(from,to) do
        local k = i;
        local string = format:gsub("%$(%w)",
            function(s)
            local c = tonumber(s, 36)-9;
            if ( rowTable[k].cells[c] and rowTable[k].cells[c].frame ) then return rowTable[k].cells[c].frame.label:GetText();
            else
                return k;
            end
        end);
        table.insert(tbl, string);
    end
    return tbl;
end


function LibSpreadsheet:GenerateRow(heap)
	local struct = {
		c = {1,1,1,1},
		b = {0,0,0,0},
		f = nil,
		h = false,
		u = true,
		cells = {}
	};
	for key, col in pairs(self.columns or {}) do
		table.insert(struct.cells,
		{
			i = nil,
			c = nil,
			b = nil,
            id = nil,
			data = nil,
			u = true,
			w = 0,
			h = false
		}
		);
	end
	table.insert(heap, struct);
	return heap[#heap];
end

function LibSpreadsheet:DeregisterRow(row)
    if ( row.frame ) then
        local x = #row.cells;
        for k, cell in pairs(row.cells) do
            if ( cell.frame ~= nil ) then
                DeleteCellFrame(cell.frame);
            end
        end
        --table.wipe(row.cells);
        row.cells = {};
        row.frame:Hide();
        row.frame:SetParent(nil);
        row.frame = nil;
        row.axis = nil;
        return x;
    end
end



local function deepcopy(object)
	local lookup_table = {}
	local function _copy(object)
		if type(object) ~= "table" then
			return object
		elseif lookup_table[object] then
			return lookup_table[object]
		end
		local new_table = {}
		lookup_table[object] = new_table
		for index, value in pairs(object) do
			new_table[_copy(index)] = _copy(value)
		end
		return setmetatable(new_table, getmetatable(object))
	end
	return _copy(object)
end

--[[
LibSpreadsheet-1.0 BOOK API
]]--

function LibSpreadsheet:SetupSheet(sheet, book)
    local xbook = book;
    sheet.parent =                    function() return xbook; end;
    sheet.rowHeap =                   {}
    sheet.heapChanged =               true;
    sheet.container =                 nil;
    sheet.xAxisScrollFrame =          nil;
    sheet.yAxisScrollFrame =          nil;
	sheet.GenerateHeap =              LibSpreadsheet.GenerateHeap;
    sheet.GenerateRow  =              LibSpreadsheet.GenerateRow;
    sheet.DeregisterRow =             LibSpreadsheet.DeregisterRow;
    sheet.AddColumn =                 LibSpreadsheet.AddColumn;
    sheet.SetColumn =                 LibSpreadsheet.SetColumn;
    sheet.DeleteColumn =              LibSpreadsheet.DeleteColumn;
    sheet.ExtractLocation =           LibSpreadsheet.ExtractLocation;
    sheet.AddRow =                    LibSpreadsheet.AddRow;
    sheet.DeleteRow =                 LibSpreadsheet.DeleteRow;
    sheet.GetValue =                  LibSpreadsheet.GetValue;
    sheet.Insert =                    LibSpreadsheet.Insert;
    sheet.Delete =                    LibSpreadsheet.Delete;
    sheet.SetCellValue =              LibSpreadsheet.SetCellValue;
    sheet.StartSelect =               LibSpreadsheet.StartSelect;
    sheet.EndSelect =                 LibSpreadsheet.EndSelect;
    sheet.SetVisibleSelection =       LibSpreadsheet.SetVisibleSelection;
    sheet.UpdateSelect =              LibSpreadsheet.UpdateSelect;
    sheet.SetInteractive =            LibSpreadsheet.SetInteractive;
    sheet.SetColorRange =             LibSpreadsheet.SetColorRange;
    sheet.SetRow =                    LibSpreadsheet.SetRow;
    sheet.ClearCells =                LibSpreadsheet.ClearCells;
    sheet.SortByColumn =              LibSpreadsheet.SortByColumn;
    sheet.SetCallback =               LibSpreadsheet.SetSheetCallback;
    sheet.UpdateFrameSizes =          LibSpreadsheet.UpdateFrameSizes;
    sheet.ShowEditBox =               LibSpreadsheet.ShowEditBox;
    sheet.SaveEditBox =               LibSpreadsheet.SaveEditBox;
    sheet.SetCellSpacing =            LibSpreadsheet.SetCellSpacing;
    sheet.SetCellPadding =            LibSpreadsheet.SetCellPadding;
    sheet.GroupBy =                   LibSpreadsheet.GroupByColumn;
    sheet.UpdateHeap =                LibSpreadsheet.UpdateHeap;
    sheet.GetUniqueValues =           LibSpreadsheet.GetUniqueValues;
    sheet.GetUniqueValuesMacro =      LibSpreadsheet.GetUniqueValuesMacro;
    sheet.AddFilter =                 LibSpreadsheet.AddFilter;
    sheet.ClearFilters =              LibSpreadsheet.ClearFilters;
    sheet.HideColumn =                LibSpreadsheet.HideColumn;
    sheet.ShowColumn =                LibSpreadsheet.ShowColumn;
    sheet.Commit =                    LibSpreadsheet.Commit;
    sheet.Export =                    LibSpreadsheet.Export;
    sheet.ParseMacro =                LibSpreadsheet.ParseMacro;
    sheet.RunMacro =                  LibSpreadsheet.RunMacro;
    sheet.SumCells =                  LibSpreadsheet.SumCells;
    sheet.ListCells =                 LibSpreadsheet.ListCells;
    for k, v in pairs (sheet.columns) do
        v.frame = nil;
        v.axis = nil;
    end
    for k, v in pairs (sheet.rows) do
        v.frame = nil;
        v.axis = nil;
        for k, c in pairs (v.cells) do
            c.frame = nil;
            c.u = true;
        end
    end
end

function LibSpreadsheet:Load(obj)
    obj.Add =               LibSpreadsheet.AddSheet;
	obj.Delete =            LibSpreadsheet.DeleteSheet;
	obj.Select =            LibSpreadsheet.Select;
	obj.StrictMode =        LibSpreadsheet.StrictMode;
    obj.ShowSelector =      LibSpreadsheet.ShowSelector;
    obj.ShowScrollbars =    LibSpreadsheet.ShowScrollbars;
    obj.Leaf =              LibSpreadsheet.Leaf;
    obj.Render =            LibSpreadsheet.Render;
    obj.Hide =              LibSpreadsheet.Hide;
    obj.BuildFramework =    LibSpreadsheet.BuildFramework;
    obj.BuildFontStrings =  LibSpreadsheet.BuildFontStrings;
    obj.SetFont =           LibSpreadsheet.SetFont;
    obj.EnableMacros =      LibSpreadsheet.EnableMacros;
    obj.ShowAxes =          LibSpreadsheet.ShowAxes;
    obj.frame =             nil;
    obj.hasBookUI =         false;
    obj.hasSelector =       false;
    obj.editbox =           nil;
    obj.headerFont =        nil;
    for k, sheet in pairs(obj.sheets) do
        LibSpreadsheet:SetupSheet(sheet, obj);
    end
end

function LibSpreadsheet:Book(title)
	local obj = {};
	if ( type(title) ~= "string" ) then error(("%s: Error in creating book title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end

    -- Book default values
	obj.sheets =            {};
    obj.parent =            self;
    obj.selectedSheet =     nil;
	obj.type =              'book';
	obj.title =             title;
    obj.strict =            false;
    obj.hasBookUI =         false;
    obj.hasSelector =       false;
    obj.editbox =           nil;

    -- Book functions
	obj.Add =               LibSpreadsheet.AddSheet;
	obj.Delete =            LibSpreadsheet.DeleteSheet;
	obj.Select =            LibSpreadsheet.Select;
	obj.StrictMode =        LibSpreadsheet.StrictMode;
    obj.ShowSelector =      LibSpreadsheet.ShowSelector;
    obj.ShowScrollbars =    LibSpreadsheet.ShowScrollbars;
    obj.Leaf =              LibSpreadsheet.Leaf;
    obj.Render =            LibSpreadsheet.Render;
    obj.Hide =              LibSpreadsheet.Hide;
    obj.BuildFramework =    LibSpreadsheet.BuildFramework;
    obj.BuildFontStrings =  LibSpreadsheet.BuildFontStrings;
    obj.SetFont =           LibSpreadsheet.SetFont;
    obj.EnableMacros =      LibSpreadsheet.EnableMacros;
    obj.ShowAxes =          LibSpreadsheet.ShowAxes;
    obj.headerFont =        nil;
    obj.numberFont =        nil;
    obj.stringFont =        nil;
	return obj;
end


--[[
LibSpreadsheet-1.0 SHEET API
]]--

function LibSpreadsheet:AddSheet(title)
	if ( type(title) ~= "string" ) then error(("%s: Error in setting sheet title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(self.sheets) do
		if ( v.title == title ) then error(("%s: Sheet \"%s\" already exists in book \"%s\"!"):format(SPREADSHEET, title, self.title)); end
	end
	local sheet = {
        parent =                    obj,
		type =                      'sheet',
		title =                     title,
        callbacks =                 {},
		columns =                   {},
        sortDirection =             LibSpreadsheet.sortAscending,
        lastColumn =                0,
		rows =                      {},
        groups =                    {},
        filters =                   {},
        hasGroups =                 false,
        hasFilters =                false,
        heapChanged =               false,
		rowHeight =                 0,
		lastRow =                   0,
		lastColumn =                0,
        currentRow =                nil,
        currentColumn =             nil,
		container =                 nil,
        isInteractive =             false,
        cellSpacing =               2,
        cellPadding =               2,
	}
    LibSpreadsheet:SetupSheet(sheet, self);
	table.insert(self.sheets, sheet);
	return self:Select(sheet.title);
end

function LibSpreadsheet:DeleteSheet(title)
    local i = nil;
	if ( type(title) ~= "string" ) then error(("%s: Error in deleting sheet; Invalid sheet name - expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(self.sheets) do
		if ( v.title == title ) then
            if ( self.selectedSheet == v ) then
                self.selectedSheet = nil;
            end
			i = k;
			break;
		end
	end
    if ( i~= nil ) then
        if ( self.sheets[i].container ~= nil ) then
            self.sheets[i].container:Hide();
            self.sheets[i].container:SetParent(nil); -- make sure the garbage collector frees up this frame and its children
            self.sheets[i].container = nil;
            self.sheets[i].xAxisScrollFrame:Hide();
            self.sheets[i].xAxisScrollFrame:SetParent(nil);
            self.sheets[i].yAxisScrollFrame:Hide();
            self.sheets[i].yAxisScrollFrame:SetParent(nil);
        end
        table.remove(self.sheets, i);
    end
end

function LibSpreadsheet:Select(sheet)
	for k, v in pairs(self.sheets) do
		if ( v.title == sheet ) then
            self.selectedSheet = v;
            if ( self.hasBookUI == true ) then
                self.verticalScrollbar:SetValue(0);
                self.horizontalScrollbar:SetValue(0);
            end
			return v;
		end
	end
	return nil;
end


--[[
LibSpreadsheet-1.0 HEADER FUNCTIONS
]]--

function LibSpreadsheet:AddColumn(...)
	for i = 1, select("#", ...) do
		local column = select(i, ...);
        if ( type(column) == "table" or type(column) == "string" ) then
            local col = {
                name =              "",
                u =                 true,
                width =             1,
                height =            1,
                c =                 {1,1,1,1},
                b =                 {0,0,0,0},
                hidden =            false,
            };
            table.insert(self.columns, col);
            if ( self:SetColumn(#self.columns, column) == false ) then
                table.remove(#self.columns);
            else
                self.lastColumn = #self.columns;
            end
        else
            error(("%s: Error while adding column, expected string or table, got %s!"):format(SPREADSHEET, column));
        end
	end
    self.heapChanged = true;
end

function LibSpreadsheet:SetColumn(column, properties)
    local c = self:ExtractLocation(column);
    if ( self.columns[c] == nil ) then error(("%s: Error while setting column %s: No such column!"):format(SPREADSHEET, LibSpreadsheet.cols[c])); end
    if ( self.columns[c].type == nil or self.columns[c].justify == nil ) then
        self.columns[c].type = "string";
        self.columns[c].justify = "LEFT";
    end
    if ( type(properties) == "string" ) then
        self.columns[c].name=properties;
        self.columns[c].u = true;
        self.heapChanged = true;
        return true;
    elseif ( type(properties) == "table" ) then
        for key, val in pairs(properties) do
            if ( key == "icon" ) then key = "i"; end
            if ( key == "color" ) then key = "c"; end
            if ( key == "background" ) then key = "b"; end
            if ( key == "texcoords" ) then key = "i"; end
            self.columns[c][key] = val;
            if ( key == "type" and properties.justify == nil ) then
                if ( val == "number" ) then
                    self.columns[c].justify = "RIGHT";
                else
                    self.columns[c].justify = "LEFT";
                end
            end
        end
        self.columns[c].u = true;
        self.heapChanged = true;
        return true;
    else
        error(("%s: Error while setting column properties, expected string or table, got %s!"):format(SPREADSHEET, type(properties)));
        return false;
    end
end


function LibSpreadsheet:DeleteColumn(column)
    local c = self:ExtractLocation(column);
    if ( self.columns[c] == nil ) then error(("%s: Error while deleting column %s: No such column!"):format(SPREADSHEET, LibSpreadsheet.cols[c])); end
    table.remove(self.columns, c);
    self.lastColumn = #self.columns;
    for k, v in pairs(self.rows) do
        if ( v.cells[c] ~= nil ) then
            if ( v.cells[c].frame ~= nil ) then
                 DeleteCellFrame(v.cells[c].frame);
             end
         end
        table.remove(v.cells, c);
    end
    self.heapChanged = true;
end



--[[
LibSpreadsheet-1.0 DATA FUNCTIONS
]]--

function LibSpreadsheet:Commit()
    if ( table.maxn(self.rowHeap) > 0 ) then
        self:ClearCells(false);
        for key, row in pairs(self.rowHeap) do
            self:DeregisterRow(row);
        end
        self.rows = self.rowHeap;
        self.rowHeap = {};
    end
end

function LibSpreadsheet:ClearCells(toggle)
    self.rows = {};
    self.hasFilters = false;
    self.hasGroups = false;
    if ( self.container ~= nil ) then
        self.container:Hide();
        self.container:SetParent(nil);
        self.container = nil;
    end
    if ( toggle ~= false ) then
        self.rowHeap = {};
    end
    self.heapChanged = true;
end

function LibSpreadsheet:GetUniqueValues(from, to)
    local tmp = {};
    local ret = {};
    local fc, fr = self:ExtractLocation(from);
    local tc, tr = self:ExtractLocation(to);
    if ( fr == nil ) then
        fr = 1;
    end
    if (tc == nil ) then
        tc = fc;
        tr = #self.rows;
    end
    local val = 0;
    if ( fr and tr and fc and tc ) then
        for x = math.min(fr, tr), math.max(fr,tr) do
            if ( self.rows[x] ~= nil ) then
                for y = math.min(fc, tc), math.max(fc, tc) do
                    if ( self.rows[x].cells[y] ~= nil and self.rows[x].cells[y].data ~= nil ) then
                        if ( tmp[self.rows[x].cells[y].data] == nil ) then
                            tmp[self.rows[x].cells[y].data] = 1;
                            table.insert(ret, self.rows[x].cells[y].data);
                        end
                    end
                end
            end
        end
    end
    return ret;
end

function LibSpreadsheet:GroupByColumn(...)
    table.wipe(self.groups);
    for i = 1, select("#", ...) do
		local column = self:ExtractLocation(select(i, ...));
        if ( self.columns[column] ~= nil ) then
            table.insert(self.groups, column);
        end
    end
    if ( table.maxn(self.groups) > 0 ) then
        self.hasGroups = true;
    else
        self.hasGroups = false;
    end
    self.heapChanged = true;
end

function LibSpreadsheet:ClearFilters()
    table.wipe(self.filters);
    self.heapChanged = true;
    self.hasFilters = false;
end

function LibSpreadsheet:AddFilter(...)
    for i = 1, select("#", ...) do
        local entry = select(i, ...);
		local column = self:ExtractLocation(entry[1]);
        if ( self.columns[column] ~= nil ) then
            table.insert(self.filters, {column, entry[2], entry[3]});
        end
    end
    self.hasFilters = true;
    self.heapChanged = true;
end


function LibSpreadsheet:SortByColumn(...)
    local func = function() return false; end;
    local i = select("#", ...);
    local p = 0;
    local N = 0;
    local rowTable = self.rows;
    self.heapChanged = true;
    if ( self.hasFilters or self.hasGroups ) then
        self:UpdateHeap();
        rowTable = self.rowHeap;
        self.heapChanged = false;
    end
    while ( i > p ) do
        local col = self:ExtractLocation(select((i-p-1), ...));
        local dir = select((i-p), ...);
        local l = string.lower;
        if      ( dir == LibSpreadsheet.sortAscending )         then dir = false;
        elseif  ( dir == LibSpreadsheet.sortAscendingWithCase ) then dir = false; l = function(x) return x; end;
        elseif  ( dir == LibSpreadsheet.sortDescending )        then dir = true;
        else                                                         dir = true;  l = function(x) return x; end;
        end
        p = p + 2;
        N = N + 1;
        local n = N;
        local r = dir;
        local c = col;
        if ( self.columns[c].type == "number" ) then
            l = function(x) return x; end;
        end
        local myFunc = func;
        func = function(a,b) if ( a == nil or b == nil ) then return; end if (l(a.cells[c].data) or 0) > (l(b.cells[c].data) or 0) then return r; elseif (l(a.cells[c].data) or 0) == (l(b.cells[c].data) or 0) then return myFunc(a,b); end return not r; end;
    end
    self.sortDirection = math.fmod(self.sortDirection, 2) + 1;
    table.sort(rowTable, func);
    for key, row in pairs(rowTable) do
        row.u = true;
    end
end


function LibSpreadsheet:AddRow(...)
    local row = self:GenerateRow(self.rows);
    for i = 1, select("#", ...) do
        local arg = select(i, ...);
        if ( type(arg) ~= "table" ) then
            arg = {data=arg};
        end
        row.cells[i].data = arg.data;
        row.cells[i].c = arg.color;
        row.cells[i].i = arg.icon;
        row.cells[i].b = arg.background;
        row.cells[i].t = arg.texcoords;
        row.cells[i].id = arg.id;
        if ( self.parent().strict == true ) then
            local t =  self.columns[i].type;
            if ( t == "string" and type(arg.data) ~= "string" ) then error(("%s: Error while adding cell in column %s, expected string, got %s!"):format(SPREADSHEET, LibSpreadsheet.cols[i], type(arg.data))); end
            if ( t == "boolean" and type(arg.data) ~= "boolean" ) then error(("%s: Error while adding cell in column %s, expected boolean, got %s!"):format(SPREADSHEET, LibSpreadsheet.cols[i], type(arg.data))); end
            if ( ( t == "number" ) and type(arg.data)  ~= "number" ) then error(("%s: Error while adding cell in column %s, expected number, got %s!"):format(SPREADSHEET, LibSpreadsheet.cols[i], type(arg.data))); end
        end
    end
    self.heapChanged = true;
    return #self.rows;
end



function LibSpreadsheet:SetRow(row, properties)
    row = tonumber(row);
    if ( self.rows[row] == nil ) then error(("%s: Error while setting row properties %s: No such row!"):format(SPREADSHEET, LibSpreadsheet.cols[c])); end
    if ( type(properties) == "table" ) then
        for key, val in pairs(properties) do
            if ( key == "color" ) then key = "c"; end
            if ( key == "background" ) then key = "b"; end
            self.rows[row][key] = val;
        end
        for k, v in pairs(self.rows[row].cells) do
            v.u = true;
        end
        self.rows[row].u = true;
    end
    self.heapChanged = true;
end

function LibSpreadsheet:DeleteRow(...)
    local dels = {};
    for i = 1, select("#", ...) do
		local row = select(i, ...);
        if ( self.rows[row] ~= nil ) then
            table.insert(dels, row);
            self:DeregisterRow(row);
        end
    end
    table.sort(dels, function(a,b) return (a > b); end);
    for k, v in pairs(dels) do
        table.remove(self.rows, v);
    end
    for k, row in pairs(self.rows) do
        row.positionChanged = true;
    end
    self.heapChanged = true;
end

function LibSpreadsheet:SetCellValue(c, r, d)
    local column, row = self:ExtractLocation(c);
    if ( row ~= nil ) then
        d = r;
    else
        row = r;
    end
    if ( self.rows[row] ~= nil ) then
        if ( self.rows[row].cells[column] ~= nil ) then
            self.rows[row].cells[column].data = d;
            self.rows[row].cells[column].u = true;
        end
    end
    self.heapChanged = true;
end

function LibSpreadsheet:ExtractLocation(str)
    local col = nil;
    local row = nil;
    if ( type(str) == "string" ) then
        local c, r = (str):match("(%D+)(%d*)");
        if ( r ~= nil ) then
            row = r;
        end
        if ( string.len(c) == 1 ) then
            col = tonumber(c, 36) - 9;
        else
            for k, column in pairs(self.columns) do
                if ( column.name == c ) then
                    col = k;
                    break;
                end
            end
        end
    elseif ( type(str) == "number" ) then
        if ( str > 26 ) then str = math.fmod(str, 26); end
        if ( str < 1 ) then
            while ( str < 1 ) do
                str = str + #self.columns;
            end
        end
        col = str;
    end
    return col, tonumber(row);
end

function LibSpreadsheet:SumCells(...)
    local fc, fr = self:ExtractLocation(select(1, ...));
    local tc, tr = self:ExtractLocation(select(2, ...));
    if ( fr == nil ) then
        fr = 1;
    end
    if (tc == nil ) then
        tc = fc;
        tr = #self.rows;
    end
    local val = 0;
    if ( tr == nil ) then
        for i = 1, select('#', ...) do
            local x = select(i, ...);
            val = val + (tonumber(x) or 0);
        end
        return val;
    end
    for x = math.min(fr, tr), math.max(fr,tr) do
        for y = math.min(fc, tc), math.max(fc, tc) do
            val = val + (tonumber(self.rows[x].cells[y].data) or 0);
        end
    end
    return val;
end

function LibSpreadsheet:ListCells(from, to)
    local fc, fr = self:ExtractLocation(from);
    local tc, tr = self:ExtractLocation(to);
    if ( fr == nil ) then
        fr = 1;
    end
    if (tc == nil ) then
        tc = fc;
        tr = #self.rows;
    end
    local vals = {};
    for x = math.min(fr, tr), math.max(fr,tr) do
        for y = math.min(fc, tc), math.max(fc, tc) do
            if ( self.rows[x] and self.rows[x].cells[y]) then
                table.insert(vals, self.rows[x].cells[y].data);
            end
        end
    end
    return unpack(vals);
end

function LibSpreadsheet:GetValue(column, row)
	local c, r = self:ExtractLocation(column);
    if ( type(r) == "number" ) then
        row = r;
    end
    if ( self.columns[c] == nil ) then error(("%s: Error in GetValue(): No such column '%s'!"):format(SPREADSHEET, LibSpreadsheet.cols[c or 27])); end
    if ( self.rows[row] == nil ) then error(("%s: Error in GetValue(): No such row '%u'!"):format(SPREADSHEET, row)); end
    return self.rows[row].cells[c].data;
end

function LibSpreadsheet:Delete(fromCol, fromRow, toCol, toRow)
    local coords = {};
    if ( type(fromCol) == "number" ) then
        coords.selectStartColumn = fromCol;
        coords.selectStartRow = fromRow;
    elseif ( type(fromCol) == "string" ) then
        local c, r = self:ExtractLocation(fromCol);
        if ( type(r) == "number" ) then
            coords.selectStartColumn = c;
            coords.selectStartRow = r;
            toRow = toCol;
            toCol = fromRow;
        else
            coords.selectStartColumn = c;
            coords.selectStartRow = fromRow;
        end
    end
    if ( type(toCol) == "number" ) then
        coords.selectEndColumn = toCol;
        coords.selectEndRow = toRow;
    elseif ( type(toCol) == "string" ) then
        local c, r = self:ExtractLocation(toCol);
        if ( type(r) == "number" ) then
            coords.selectEndColumn = c;
            coords.selectEndRow = r;
        else
            coords.selectEndColumn = c;
            coords.selectEndRow = toRow;
        end
    end
    for x = math.min(coords.selectStartRow, coords.selectEndRow), math.max(coords.selectStartRow, coords.selectEndRow) do
        for y = math.min(coords.selectStartColumn, coords.selectEndColumn), math.max(coords.selectStartColumn, coords.selectEndColumn) do
            self.rows[x].cells[y].data = nil;
        end
    end
    self.heapChanged = true;
end

function LibSpreadsheet:Insert(...)

end

function LibSpreadsheet:EnableMacros(toggle)
	if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting macro mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.macros = toggle;
end

function LibSpreadsheet:StrictMode(toggle)
	if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting strict mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.strict = toggle;
end

function LibSpreadsheet:UpdateHeap()
    if ( self.heapChanged ) then
        local x = 0;
        for key, row in pairs(self.rowHeap) do
            x = x + (self:DeregisterRow(row) or 0) + 1;
        end
--        print("Cleared " .. x .. " frames.");
        local rowTable = self.rows;
        if ( self.hasFilters ) then
            for key, entry in pairs(self.filters) do
                local tmp = {};
                local col = entry[1];
                local method = entry[2];
                local val = entry[3];
                for key, row in pairs(rowTable) do
                    if (
                        ( method == "=" and  row.cells[col].data == val ) or
                        ( method == "~=" and row.cells[col].data ~= val ) or
                        ( method == ">" and  row.cells[col].data >  val ) or
                        ( method == "<" and  row.cells[col].data <  val ) or
                        ( method == ">=" and row.cells[col].data >= val ) or
                        ( method == "<=" and row.cells[col].data <= val )
                    ) then
                        local r = deepcopy(row);
                        r.frame = nil;
                        r.u = true;
                        for k, v in pairs(r.cells) do v.frame = nil; v.u=true; end
                        table.insert(tmp, r);
                    end
                end
                rowTable = tmp;
            end
        end
        if ( self.hasGroups ) then
            local x = #rowTable;
            for key, col in pairs(self.groups) do
                local tmp = {};
                for key, row in pairs(rowTable) do
                    if ( tmp[row.cells[col].data] ~= nil ) then
                        for key, cell in pairs(row.cells) do
                            if ( self.columns[key].type == "number" and key ~= column ) then
                                tmp[row.cells[col].data].cells[key].data = (tmp[row.cells[col].data].cells[key].data or 0) + (cell.data or 0);
                            end
                        end
                    else
                        tmp[row.cells[col].data] = {cells={}};
                        for k, cell in pairs(row.cells) do
                            tmp[row.cells[col].data].cells[k] = deepcopy(cell);
                            tmp[row.cells[col].data].cells[k].frame = nil;
                            tmp[row.cells[col].data].cells[k].u = true;
                        end
                    end
                end
                local ntmp = {};
                for k, v in pairs(tmp) do
                    table.insert(ntmp, v);
                end
                table.wipe(tmp);
                rowTable = ntmp;
            end
            local y = #rowTable;
        end
        self.rowHeap = rowTable;
        self.heapChanged = false;
    end
end



--[[
LibSpreadsheet-1.0 VISUAL HANDLERS
]]--


function LibSpreadsheet:ShowSelector(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting selector mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.showSelector = toggle;
end

function LibSpreadsheet:ShowScrollbars(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting scrollbar visibility; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.showScrollbars = toggle;
end

function LibSpreadsheet:SetInteractive(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting interactive mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
    self.isInteractive = toggle;
end

function LibSpreadsheet:UpdateFrameSizes()
    local tex = nil;
    if ( UIParent ) then tex = UIParent:CreateTexture(nil, "BACKGROUND"); end
    local rowTable = self.rows;
    if ( self.hasGroups or self.hasFilters ) then
        self:UpdateHeap();
        rowTable = self.rowHeap;
    end
    for key, header in pairs (self.columns) do
        if ( header.u ) then
            self.parent().str:SetFontObject(self.parent().headerFont);
            self.parent().str:SetText((header.name or "??") .. ":");
            local height = math.ceil(self.parent().str:GetStringHeight());
            local width = math.ceil(self.parent().str:GetStringWidth());
            if ( header.h ~= height ) then header.h = height; end
            if ( header.w ~= width ) then header.w = width; end
            self.rowHeight = height + (self.cellPadding*2);
            header.colwidth = width + (self.cellPadding*2);
        end
    end
    local t = nil;
    for key, row in pairs (rowTable) do
        for key, cell in pairs (row.cells) do
            t = self.columns[key].type;
            if ( cell.u or cell.macroFunction ) then
                if ( cell.u ) then
                    if ( type(cell.data) == "string" and cell.data:sub(1,1) == "=" and self.parent().macros == true ) then
                        cell.macroFunction = loadstring("return (" ..self:ParseMacro(cell.data:sub(2), false) .. ") ");
                    else
                        cell.macroFunction = nil;
                    end
                end
                if ( cell.macroFunction ) then
                    cell.macroData = cell.macroFunction()
                end
                if ( t == "number" ) then
                    self.parent().str:SetFontObject(self.parent().numberFont);
                    cell.data = tonumber(cell.macroData or cell.data);
                    local s = ("%%.%uf%%s   "):format(self.columns[key].decimals or 0);
                    self.parent().str:SetText(string.format(s, cell.macroData or cell.data or 0, (self.columns[key].suffix or "")));
                else
                    self.parent().str:SetFontObject(self.parent().stringFont);
                    self.parent().str:SetText(cell.macroData or cell.data or "");
                end
                local cw, ch = 0, 0;
                cell.h = math.ceil(self.parent().str:GetStringHeight());
                cell.w = math.ceil(self.parent().str:GetStringWidth() + (self.cellPadding*2) + cw) + 2;
                if ( cell.i ) then
                    cell.w = cell.w + self.rowHeight + self.cellPadding;
                    cell.h = self.rowHeight;
                end
            end
            if ( cell.w > self.columns[key].colwidth ) then
                self.columns[key].colwidth = cell.w;
            end
            if ( cell.h > self.rowHeight ) then
                self.rowHeight = cell.h + (self.cellPadding*2);
            end
        end
    end
end

function LibSpreadsheet:BuildFontStrings()
    self.str = UIParent:CreateFontString();
    local garble = string.format("%8x", math.random(time()));
    self.headerFont = CreateFont("LibSpreadsheetHeaderFont"..garble);
    self.numberFont = CreateFont("LibSpreadsheetNumberFont"..garble);
    self.stringFont = CreateFont("LibSpreadsheetStringFont"..garble);
    local locale =            (GetLocale or (function() return "enUS"; end))();
    self.numberFont:SetFont(    [[Fonts\ARIALN.TTF]],   12);
    if ( locale == "zhCN" ) then
        self.headerFont:SetFont([[Fonts\ZYKai_T.TTF]],  12);
        self.stringFont:SetFont([[Fonts\ZYKai_T.TTF]],  10);
    elseif ( locale == "zhTW" ) then
        self.headerFont:SetFont([[Fonts\bLEI00D.TTF]],  12);
        self.stringFont:SetFont([[Fonts\bLEI00D.TTF]],  10);
    elseif ( locale == "koKR" ) then
        self.headerFont:SetFont([[Fonts\2002.TTF]],     12);
        self.stringFont:SetFont([[Fonts\2002.TTF]],     10);
    else
        self.headerFont:SetFont([[Fonts\FRIZQT__.TTF]], 12);
        self.stringFont:SetFont([[Fonts\FRIZQT__.TTF]], 10);
    end
end

function LibSpreadsheet:SetFont(what, font, size, flags)
    what = string.lower(what);
    local garble = string.format("%8x", math.random(time()));
    local fontstring = CreateFont("LibSpreadsheetCustomFont"..garble);
    fontstring:SetFont(font, size, flags);
    if ( what == "header") then self.headerFont = fontstring end
    if ( what == "string") then self.headerFont = fontstring end
    if ( what == "number") then self.headerFont = fontstring end
end

function LibSpreadsheet:BuildFramework()
    if ( self.headerFont == nil ) then
        self:BuildFontStrings();
    end
    self.frame = CreateFrame("Frame", nil, UIParent);
    self.outerScrollframe = CreateFrame("ScrollFrame", nil, self.frame);
    self.header = nil;
    self.frame:SetPoint("CENTER", UIParent, "CENTER", 0,0);
    self.frame:SetFrameStrata("DIALOG")
    self.frame:EnableMouse(true)
    self.innerFrame = nil;
    self.frame:Hide();
    self.offsetY = 0;
    self.scrollFrame = CreateFrame("ScrollFrame", nil, self.frame);
	self.scrollFrame:SetPoint("TOPLEFT");
	self.scrollFrame:SetPoint("BOTTOMRIGHT");
	self.scrollFrame:EnableMouseWheel(true)
    self.scrollFrame:EnableMouse(true)
    self.scrollFrame:SetAllPoints();

    self.verticalScrollbar = CreateFrame("Slider", nil, self.scrollFrame, "UIPanelScrollBarTemplate");
    self.verticalScrollbar:SetOrientation('VERTICAL');
	self.verticalScrollbar:SetPoint("TOPLEFT", self.scrollFrame, "TOPRIGHT", 4, -16)
	self.verticalScrollbar:SetPoint("BOTTOMLEFT", self.scrollFrame, "BOTTOMRIGHT", 4, 16)
	self.verticalScrollbar:SetMinMaxValues(0, 1000)
	self.verticalScrollbar:SetValueStep(1)
	self.verticalScrollbar:SetValue(0)
    self.verticalScrollbar:SetWidth(16)
    self.scrollFrame.scrollvalueY = 0;
    self.scrollFrame.scrollvalueX = 0;
	self.verticalScrollbar:Hide()
	self.verticalScrollbar:SetScript("OnValueChanged", function(a,b) self.scrollFrame:scrollFunc(b,nil); end );

    local s2 = sqrt(2);
    local cos, sin, rad = math.cos, math.sin, math.rad;
    local function CalculateCorner(angle)
        local r = rad(angle);
        return 0.5 + cos(r) / s2, 0.5 + sin(r) / s2;
    end
    local function RotateTexture(texture, angle)
        local LRx, LRy = CalculateCorner(angle + 45);
        local LLx, LLy = CalculateCorner(angle + 135);
        local ULx, ULy = CalculateCorner(angle + 225);
        local URx, URy = CalculateCorner(angle - 45);

        texture:SetTexCoord(ULx, ULy, LLx, LLy, URx, URy, LRx, LRy);
    end


    local garble = string.format("%8x", math.random(time()));
    self.horizontalScrollbar = CreateFrame("Slider", "LibSpreadsheet_Scroller_" .. garble, self.scrollFrame, "UIPanelScrollBarTemplate");
    self.horizontalScrollbar:SetOrientation('HORIZONTAL');
    self.horizontalScrollbar:SetPoint("TOPLEFT", self.scrollFrame, "BOTTOMLEFT", 16, -5)
	self.horizontalScrollbar:SetPoint("BOTTOMRIGHT", self.scrollFrame, "BOTTOMRIGHT", 0, -21)
	self.horizontalScrollbar:SetMinMaxValues(0, 1000)
	self.horizontalScrollbar:SetValueStep(1)
	self.horizontalScrollbar:SetValue(0)
	self.horizontalScrollbar:Hide()
	self.horizontalScrollbar:SetScript("OnValueChanged", function(a,b) self.scrollFrame:scrollFunc(nil,b); end );
    local tex, button;
    button = _G["LibSpreadsheet_Scroller_" .. garble .. "ScrollUpButton"];
    tex = button:GetNormalTexture(); RotateTexture(tex, 90); tex:SetPoint("TOPLEFT", button, "TOPLEFT", -8, 8); tex:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 8, -8);
    tex = button:GetPushedTexture(); RotateTexture(tex, 90); tex:SetPoint("TOPLEFT", button, "TOPLEFT", -8, 8); tex:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 8, -8);
    tex = button:GetHighlightTexture(); RotateTexture(tex, 90); tex:SetPoint("TOPLEFT", button, "TOPLEFT", -8, 8); tex:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 8, -8);
    button:ClearAllPoints();
    button:SetPoint("BOTTOMRIGHT", self.horizontalScrollbar, "BOTTOMLEFT", 0, 0);
    --button:SetPoint("TOPRIGHT", self.horizontalScrollbar, "TOPLEFT", 0, 0);

    button = _G["LibSpreadsheet_Scroller_" .. garble .. "ScrollDownButton"];
    tex = button:GetNormalTexture(); RotateTexture(tex, 90); tex:SetPoint("TOPLEFT", button, "TOPLEFT", -8, 8); tex:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 8, -8);
    tex = button:GetPushedTexture(); RotateTexture(tex, 90); tex:SetPoint("TOPLEFT", button, "TOPLEFT", -8, 8); tex:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 8, -8);
    tex = button:GetHighlightTexture(); RotateTexture(tex, 90); tex:SetPoint("TOPLEFT", button, "TOPLEFT", -8, 8); tex:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 8, -8);
    button:ClearAllPoints();
    button:SetPoint("TOPLEFT", self.horizontalScrollbar, "TOPRIGHT", 0, 0);

    self.scrollFrame.onMouseWheel = function(caller, value)
		local height, viewheight = self.scrollFrame:GetHeight(), self.innerFrame:GetHeight()

		if height > viewheight then
			self.verticalScrollbar:Hide()
		else
            self.verticalScrollbar:Show()
			local diff = height - viewheight
			local delta = 1
			if (value or 0) < 0 then
				delta = -1
			end
			self.verticalScrollbar:SetValue(min(max(self.scrollFrame.scrollvalueY + delta*(self.offsetY/(diff/45)),0), 1000))
		end
	end;


    self.sheetSelector = CreateFrame("Frame", nil, self.frame);
    self.sheetSelector:SetFrameStrata("FULLSCREEN_DIALOG")
    self.prevSheet = CreateFrame("Button", nil, self.sheetSelector);
    self.prevSheet:SetWidth(32);
    self.prevSheet:SetHeight(32);
    self.prevSheet:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up");
    self.prevSheet:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down");
    self.prevSheet:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled");
    self.prevSheet:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD");
    self.prevSheet:SetScript("OnClick", function() self:Leaf(-1); self:Render(); end);
    self.prevSheet:SetPoint("LEFT", self.sheetSelector, "LEFT", 0, 0);


    self.nextSheet = CreateFrame("Button", nil, self.sheetSelector);
    self.nextSheet:SetWidth(32);
    self.nextSheet:SetHeight(32);
    self.nextSheet:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up");
    self.nextSheet:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down");
    self.nextSheet:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled");
    self.nextSheet:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD");
    self.nextSheet:SetScript("OnClick", function() self:Leaf(1); self:Render(); end);
    self.nextSheet:SetPoint("RIGHT", self.sheetSelector, "RIGHT", 0, 0);

    self.sheetSelector:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT", 0, 0);
    self.sheetSelector:SetPoint("TOPRIGHT", self.frame, "BOTTOMRIGHT", 0, 32);

    self.sheetSelectorStatus = CreateFrame("Frame", nil, self.sheetSelector);
	self.sheetSelectorStatus:SetBackdrop({
        bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
        edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
        tile = true, tileSize = 16, edgeSize = 16,
        insets = { left = 3, right = 3, top = 5, bottom = 3 }
    });
    self.sheetSelectorStatus:SetAllPoints();
	self.sheetSelectorStatus:SetBackdropColor(0.1, 0.1, 0.1, 0.5);
	self.sheetSelectorStatus:SetBackdropBorderColor(0.4, 0.4, 0.4);

    self.sheetSelectorText = self.sheetSelectorStatus:CreateFontString(nil, "DIALOG");
    self.sheetSelectorText:SetFontObject(self.headerFont);
    self.sheetSelectorText:SetAllPoints(self.sheetSelectorStatus);
    self.sheetSelectorText:SetText("");
    self.sheetSelectorText:SetTextColor(0.9,0.75,0.2);
    self.sheetSelectorText:Show();
	self.scrollFrame:SetScript("OnMouseWheel", self.scrollFrame.onMouseWheel);
    self.outerScrollframe:SetScript("OnShow", function() self.scrollFrame:onMouseWheel(0); end);
    self.scrollFrame.scrollFunc = function(caller, valueY, valueX)
		local viewheight = self.scrollFrame:GetHeight()
        local viewwidth = self.scrollFrame:GetWidth()
		local height = self.innerFrame:GetHeight()
        local width = self.innerFrame:GetWidth()
		local offsetX, offsetY

		if viewheight > height then
			offsetY = 0
            self.verticalScrollbar:Hide()
		else
            self.verticalScrollbar:SetMinMaxValues(0, height-viewheight)
            self.verticalScrollbar:Show()
			offsetY = valueY or self.scrollFrame.scrollvalueY;
		end
        if viewwidth >= width then
			offsetX = 0
            self.horizontalScrollbar:Hide()
		else
            self.horizontalScrollbar:SetMinMaxValues(0, width-viewwidth)
            self.horizontalScrollbar:Show()
			offsetX = valueX or self.scrollFrame.scrollvalueX;
		end
		self.innerFrame:ClearAllPoints()
		self.innerFrame:SetPoint("TOPLEFT", -offsetX, offsetY)
		self.innerFrame:SetPoint("TOPRIGHT", -offsetX, offsetY)
        self.selectedSheet.yAxis:ClearAllPoints()
        self.selectedSheet.yAxis:SetPoint("TOPLEFT", 0, offsetY)
        self.selectedSheet.xAxis:SetPoint("TOPLEFT", -offsetX, 0)

		self.scrollFrame.scrollvalueY = valueY or self.scrollFrame.scrollvalueY;
        self.scrollFrame.scrollvalueX = valueX or self.scrollFrame.scrollvalueX;
    end;

    self.editbox = CreateFrame("EditBox", "DIALOG", self.frame, "InputBoxTemplate");
	self.editbox:SetAutoFocus(true);
	self.editbox:SetFontObject(self.stringFont);
	self.editbox:SetScript("OnEscapePressed", function(me) me:Hide(); end);
    self.editbox:SetScript("OnHide", function(me) if ( me.caller ~= nil ) then me.caller:Show();end end);
	self.editbox:SetScript("OnEnterPressed",  function(me) if ( me.caller ~= nil ) then me.caller.sheet:SaveEditBox(me.caller, me:GetText()); me:Hide(); end end);
	self.editbox:SetTextInsets(0, 0, 3, 3);
	self.editbox:SetMaxLetters(256);
    self.editbox:Hide();
    self.hasBookUI = true;
end

function LibSpreadsheet:Hide()
    if ( self.frame ~= nil ) then
        self.frame:Hide();
    end
end

function LibSpreadsheet:Render()
    if ( self.hasBookUI == false ) then
        self:BuildFramework();
    end
    local sheet = self.selectedSheet;
    sheet:UpdateFrameSizes();
    for k, s in pairs(self.sheets) do
        if ( s.container ~= nil and s.container.Hide ~= nil ) then
            s.container:Hide();
            s.xAxisScrollFrame:Hide();
            s.yAxisScrollFrame:Hide();
        end
    end

    if ( self.innerFrame ~= nil ) then self.innerFrame:Hide(); end

    --[[ SHEET AXES AND DATA CONTAINER PROTOTYPES ]]--
    if ( sheet.container == nil ) then
        sheet.xAxisScrollFrame = CreateFrame("ScrollFrame", nil, self.frame);
        sheet.yAxisScrollFrame = CreateFrame("ScrollFrame", nil, self.frame);

        sheet.xAxisScrollFrame:SetPoint("TOPRIGHT", self.frame, "TOPRIGHT", 0, 0);
        sheet.xAxisScrollFrame:SetPoint("BOTTOMLEFT", sheet.yAxisScrollFrame, "TOPRIGHT", 0, 0);

        sheet.yAxisScrollFrame:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT", 0, 16);
        sheet.yAxisScrollFrame:SetPoint("TOPLEFT", self.frame, "TOPLEFT", 0, 0);
        sheet.yAxisScrollFrame:SetWidth(20);

        sheet.xAxis = CreateFrame("Frame", nil, sheet.xAxisScrollFrame);
        sheet.xAxis:SetPoint("TOPLEFT", sheet.xAxisScrollFrame, "TOPLEFT", 0, 0);
        sheet.xAxisScrollFrame:SetScrollChild(sheet.xAxis);

        sheet.yAxis = CreateFrame("Frame", nil, sheet.yAxisScrollFrame);
        sheet.yAxis:SetPoint("TOPLEFT", sheet.yAxisScrollFrame, "TOPLEFT", 0, 0);
        sheet.yAxisScrollFrame:SetScrollChild(sheet.yAxis);

        sheet.container = CreateFrame("Frame", nil, self.scrollFrame);
        sheet.container:SetPoint("TOPLEFT", self.scrollFrame, "TOPLEFT", 0, 0);
    end

    --[[ SET UP X AXIS ]]--
    local offsetX = 0;

    for index, column in pairs(sheet.columns) do
        if ( not column.hidden ) then
            if ( self.showAxes ) then
                sheet.xAxisScrollFrame:SetHeight(2 * (sheet.rowHeight + sheet.cellSpacing));
                sheet.xAxis:SetHeight(2 * (sheet.rowHeight + sheet.cellSpacing));
                if ( column.axis == nil ) then
                    column.axis = CreateCellFrame("Button", nil, sheet.xAxis);
                    column.axis.label:SetJustifyH("CENTER");
                    column.axis.label:SetFontObject(self.headerFont);
                    column.axis.label:SetPoint("TOPLEFT", column.axis, "TOPLEFT", (sheet.cellPadding), -(sheet.cellPadding));
                    column.axis.label:SetPoint("BOTTOMRIGHT", column.axis, "BOTTOMRIGHT", -(sheet.cellPadding), (sheet.cellPadding));
                    column.axis.label:SetText(LibSpreadsheet.cols[index]);
                    column.axis.texture = column.axis:CreateTexture();
                    column.axis.texture:SetTexture(0.2,0.2,0.4,0.5);
                    column.axis.texture:SetAllPoints(column.axis);
                end
            else
                sheet.xAxisScrollFrame:SetHeight(1 * (sheet.rowHeight + sheet.cellSpacing));
                sheet.xAxis:SetHeight(1 * (sheet.rowHeight + sheet.cellSpacing));
            end
            if ( column.frame == nil ) then
                column.frame = CreateCellFrame("Button", nil, sheet.xAxis);
                column.frame.label:SetJustifyH(column.justify or "LEFT");
                column.frame.label:SetFontObject(self.headerFont);
                column.frame.label:SetPoint("TOPLEFT", column.frame, "TOPLEFT", (sheet.cellPadding), -(sheet.cellPadding));
                column.frame.label:SetPoint("BOTTOMRIGHT", column.frame, "BOTTOMRIGHT", -(sheet.cellPadding), (sheet.cellPadding));
                column.frame.label:SetText(column.name);
                column.frame.label:SetTextColor(0.9,0.75,0.2);
            end
            if ( column.u ) then
                if ( self.showAxes ) then
                    column.axis:SetPoint("TOPLEFT", sheet.xAxis, "TOPLEFT", offsetX, 0);
                    column.axis:SetWidth(column.colwidth);
                    column.axis:SetHeight(sheet.rowHeight);
                    column.axis:SetScript("OnMouseDown", function(caller, event) sheet:StartSelect({column=true,index=index}, event); end);
                    column.axis:SetScript("OnMouseUp", function(caller, event) sheet:EndSelect({column=true,index=index}, event); end);
                    column.axis:SetScript("OnEnter", function(caller, event) sheet:UpdateSelect({column=true,index=index}); end);
                    column.axis:SetScript("OnLeave", function(caller, event) sheet:UpdateSelect({column=true,index=index}); end);
                end
                column.frame:SetPoint("BOTTOMLEFT", sheet.xAxis, "BOTTOMLEFT", offsetX, 0);
                column.frame:SetWidth(column.colwidth);
                column.frame:SetHeight(sheet.rowHeight);
                if ( column.sort == true ) then
                    local k = index;
                    column.frame:SetScript("OnClick",
                        function()
                            local l = true;
                            if ( sheet.callbacks["OnUserSortRequested"] ~= nil ) then
                                if ( type(sheet.callbacks["OnUserSortRequested"]) == "function" ) then
                                    l = (sheet.callbacks["OnUserSortRequested"])(LibSpreadsheet.cols[k]);
                                end
                            end
                            if ( l ) then
                                sheet:SortByColumn(k, sheet.sortDirection);
                                self:Render();
                                if ( sheet.callbacks["OnUserSortProcessed"] ~= nil ) then
                                    if ( type(sheet.callbacks["OnUserSortProcessed"]) == "function" ) then
                                        (sheet.callbacks["OnUserSortProcessed"])(LibSpreadsheet.cols[k]);
                                    end
                                end
                            end
                        end
                    );
                end
            end
            offsetX = offsetX + column.colwidth + sheet.cellSpacing;
        end
    end
    sheet.xAxis:SetWidth(offsetX);


    --[[ SET UP ROWS AND Y AXIS ]]--

    offsetX = 0;
    local offsetY = 0;

    if ( self.showAxes ) then
        sheet.yAxisScrollFrame:SetWidth(30);
        sheet.yAxis:SetWidth(30);
    else
        sheet.yAxisScrollFrame:SetWidth(1);
        sheet.yAxis:SetWidth(1);
    end

    local rowTable = sheet.rows;
    if ( sheet.hasFilters or sheet.hasGroups ) then
        rowTable = sheet.rowHeap;
    end
    local f = 0;
    for index, row in pairs(rowTable) do
        if ( self.showAxes ) then
            if ( row.axis == nil ) then
                row.axis = CreateCellFrame("Button", nil, sheet.yAxis);
                row.axis.label:SetJustifyH("CENTER");
                row.axis.label:SetFontObject(self.headerFont);
                row.axis.label:SetPoint("TOPLEFT", row.axis, "TOPLEFT", (sheet.cellPadding), -(sheet.cellPadding));
                row.axis.label:SetPoint("BOTTOMRIGHT", row.axis, "BOTTOMRIGHT", -(sheet.cellPadding), (sheet.cellPadding));
                row.axis.texture = row.axis:CreateTexture();
                row.axis.texture:SetTexture(0.2,0.2,0.4,0.5);
                row.axis.texture:SetAllPoints(row.axis);
            end
        end
        if ( row.frame == nil ) then
            row.frame = CreateFrame("Frame", nil, sheet.container);
            row.frame.texture = row.frame:CreateTexture();
            row.frame.texture:SetAllPoints(row.frame);
             f = f + 1;
        end
        if ( row.u ) then
            if ( row.axis ) then
                row.axis:SetPoint("TOPLEFT", sheet.yAxis, "TOPLEFT", 0, -offsetY);
                row.axis:SetWidth(30);
                row.axis:SetHeight(sheet.rowHeight);
                row.axis:SetHeight(sheet.rowHeight);
                row.axis.label:SetText(index);
                row.axis:SetScript("OnMouseDown", function(caller, event) sheet:StartSelect({row=true,index=index}, event); end);
                row.axis:SetScript("OnMouseUp", function(caller, event) sheet:EndSelect({row=true,index=index}, event); end);
                row.axis:SetScript("OnEnter", function(caller, event) sheet:UpdateSelect({row=true,index=index}); end);
                row.axis:SetScript("OnLeave", function(caller, event) sheet:UpdateSelect({row=true,index=index}); end);
            end
            if ( row.b ~= nil ) then
                row.frame.texture:SetTexture(row.b.r or row.b[1], row.b.g or row.b[2], row.b.b  or row.b[3], row.b.a or row.b[4]);
            end
            row.frame:SetPoint("TOPLEFT", sheet.container, "TOPLEFT", 0, -offsetY);
            row.frame:SetHeight(sheet.rowHeight);
        end
        offsetY = offsetY + sheet.rowHeight + sheet.cellSpacing;

        --[[ SET UP CELLS ]]--
        offsetX = 0;
        for cellIndex, cell in pairs(row.cells) do
            f = f + 1;
            if ( not sheet.columns[cellIndex].hidden ) then
                if ( cell.frame == nil ) then
                    cell.frame = CreateCellFrame("Button", nil, row.frame);
                    cell.frame.icon:SetPoint("TOPLEFT", cell.frame, "TOPLEFT", (sheet.cellPadding), 0);
                    cell.frame.icon:SetTexCoord(0, 1, 0, 1);
                    cell.frame.label:SetPoint("TOPLEFT", cell.frame.icon, "TOPRIGHT", (sheet.cellPadding), -(sheet.cellPadding));
                    cell.frame.label:SetPoint("BOTTOMRIGHT", cell.frame, "BOTTOMRIGHT", -(sheet.cellPadding), (sheet.cellPadding));
                    cell.frame.overlay:SetAllPoints();
                    cell.frame.overlay:SetBlendMode("ADD");
                    cell.frame.overlay:Hide();
                    cell.frame.background:SetAllPoints();
                    cell.frame.background:Hide();
                    cell.frame.column = cellIndex;
                    cell.frame.sheet = sheet;
                    cell.frame.cell = cell;

                end
                if ( cell.u ) then
                    if ( cell.i ) then
                        cell.frame.icon:SetTexture(cell.i);
                        if ( cell.t ) then
                            cell.frame.icon:SetTexCoord(cell.t[1], cell.t[2], cell.t[3], cell.t[4]);
                        end
                        cell.frame.icon:SetHeight(sheet.rowHeight);
                        cell.frame.icon:SetWidth(sheet.rowHeight);
                        cell.frame.icon:Show();
                    else
                        cell.frame.icon:SetTexture(nil);
                        cell.frame.icon:Hide();
                        cell.frame.icon:SetHeight(0.1);
                        cell.frame.icon:SetWidth(0.1);
                    end
                    cell.frame:SetPoint("TOPLEFT", row.frame, "TOPLEFT", offsetX, -(sheet.cellSpacing/2));
                    cell.frame:SetWidth(sheet.columns[cellIndex].colwidth);
                    cell.frame:SetHeight(sheet.rowHeight);

                    cell.frame.row = index;
                    cell.frame.label:SetJustifyH(sheet.columns[cellIndex].justify);
                    if ( cell.c == nil ) then cell.c = {}; end
                    cell.frame.label:SetTextColor((cell.c.r or cell.c[1]) or sheet.columns[cellIndex].c[1], (cell.c.g or cell.c[2]) or sheet.columns[cellIndex].c[2], (cell.c.b or cell.c[3]) or sheet.columns[cellIndex].c[3], 1);
                    cell.frame.background:SetTexture(sheet.columns[cellIndex].b[1], sheet.columns[cellIndex].b[2], sheet.columns[cellIndex].b[3], sheet.columns[cellIndex].b[4] or 1);

                    if ( sheet.isInteractive or sheet.columns[cellIndex].interactive ) then
                        cell.frame:SetScript("OnMouseDown", function(caller, event) sheet:StartSelect(cell.frame, event); end);
                        cell.frame:SetScript("OnMouseUp", function(caller, event) sheet:EndSelect(cell.frame, event); end);
                        cell.frame:SetScript("OnEnter", function(caller, event) sheet:UpdateSelect(cell.frame); end);
                        cell.frame:SetScript("OnLeave", function(caller, event) sheet:UpdateSelect(cell.frame); end);
                        cell.frame:SetScript("OnDoubleclick", function(caller, event) sheet:ShowEditBox(cell.frame); end);
                    end
                    if ( sheet.columns[cellIndex].onClick ~= nil ) then
                        local func = sheet.columns[cellIndex].onClick;
                        cell.frame:SetScript("OnClick", function(caller, event) func(cell.id, cell.data, cell.frame); end);
                    end
                    if ( sheet.columns[cellIndex].type == "number" ) then
                        cell.frame.label:SetFontObject(self.numberFont);
                        local cellData = cell.data or 0;
                        if ( cell.macroData ) then
                            cellData = cell.macroData;
                        end
                        local extra = "";
                        if ( sheet.columns[cellIndex].decimals or 0 > 0 ) then
                            local extra = string.sub(string.format ( ("%%.%uf"):format(sheet.columns[cellIndex].decimals or 0), cellData - math.floor(cellData)), 3);
                        end
                        local number = ("%.0f"):format(cellData or 0);
                        local left,num,right = string.match(number,'^([^%d]*%d)(%d+)(.-)$');
                        local prettyNum = (left and left..(num:reverse():gsub('(%d%d%d)','%1,'):reverse()) or number);
                        cell.frame.label:SetText(string.format("%s%s%s", prettyNum, extra, (sheet.columns[cellIndex].suffix or "")));
                        if ( sheet.columns[cellIndex].colorRange ~= nil ) then
                            if ( sheet.columns[cellIndex].minval ~= nil and sheet.columns[cellIndex].maxval ~= nil ) then
                                local colors = table.maxn(sheet.columns[cellIndex].colorRange);
                                if ( sheet.columns[cellIndex].maxval ~= nil and (cell.data or 0) > sheet.columns[cellIndex].maxval ) then
                                    cell.data = sheet.columns[cellIndex].maxval;
                                end
                                local where = ( ( (cellData - sheet.columns[cellIndex].minval)  / (sheet.columns[cellIndex].maxval - sheet.columns[cellIndex].minval)) * (colors-1) ) + 1;
                                local p, n = math.floor(where), math.ceil(where);
                                local r = (where - p) * 100;
                                if (  sheet.columns[cellIndex].colorRange[p] == nil ) then
                                    where = ((colors-1)/2) + 1
                                    p, n = math.floor(where), math.ceil(where);
                                    r = 0;
                                end
                                local color = {
                                    (  ( sheet.columns[cellIndex].colorRange[p][1] * (100-r) ) + ( sheet.columns[cellIndex].colorRange[n][1] * (r) ) ) / 100,
                                    (  ( sheet.columns[cellIndex].colorRange[p][2] * (100-r) ) + ( sheet.columns[cellIndex].colorRange[n][2] * (r) ) ) / 100,
                                    (  ( sheet.columns[cellIndex].colorRange[p][3] * (100-r) ) + ( sheet.columns[cellIndex].colorRange[n][3] * (r) ) ) / 100
                                    };
                                cell.frame.label:SetTextColor(color[1], color[2], color[3]);
                                if ( sheet.columns[cellIndex].dominant ) then
                                    for k, v in pairs(sheet.rows[index].cells) do
                                        v.color = {color[1], color[2], color[3]};
                                        if ( v.frame ) then
                                            v.frame.label:SetTextColor(color[1], color[2], color[3]);
                                        end
                                    end
                                end
                            end
                        end
                    elseif ( sheet.columns[cellIndex].type == "string" ) then
                        cell.frame.label:SetFontObject(self.stringFont);
                        cell.frame.label:SetText(cell.macroData or cell.data or "");
                    elseif ( sheet.columns[cellIndex].type == "boolean" ) then
                        cell.frame.label:SetFontObject(self.stringFont);
                        if ( cell.data == true ) then cell.frame.label:SetText("true");
                        else cell.frame.label:SetText("false"); end
                    end
                end
                offsetX = offsetX + sheet.columns[cellIndex].colwidth + sheet.cellSpacing;
            end
        end
        row.frame:SetWidth(offsetX);
    end
    sheet.yAxisScrollFrame:Show();
    sheet.xAxisScrollFrame:Show();
    sheet.container:Show();
    sheet.yAxis:SetHeight(offsetY);
    sheet.yAxisScrollFrame:SetPoint("TOPLEFT", self.frame, "TOPLEFT", 0, -(sheet.xAxisScrollFrame:GetHeight()));

    self.innerFrame = sheet.container;
    self.scrollFrame:SetScrollChild(sheet.container);
    self.offsetY = offsetY;
    self.scrollFrame:SetPoint("TOPLEFT", sheet.yAxisScrollFrame, "TOPRIGHT", 0, 0);
    self.scrollFrame:SetPoint("BOTTOMRIGHT", self.frame, "BOTTOMRIGHT", -16, 16);
    self.verticalScrollbar:SetMinMaxValues(0, offsetY);
    self.verticalScrollbar:SetValue(0);
    self.innerFrame:SetHeight(offsetY);
    self.innerFrame:SetWidth(offsetX);
--    self.scrollFrame:scrollFunc(0);
    --self.scrollFrame:onMouseWheel(0);
    sheet.xAxis:SetPoint("TOPLEFT", sheet.xAxisScrollFrame, "TOPLEFT", 0, 0);
    sheet.yAxis:SetPoint("TOPLEFT", 0, 0);
    self.sheetSelector:Hide();
    if ( self.showSelector == true ) then
        self:Leaf(0);
        self.scrollFrame:SetPoint("BOTTOMRIGHT", self.frame, "BOTTOMRIGHT", -16, 64);
        sheet.yAxisScrollFrame:SetPoint("BOTTOMLEFT", self.frame, "BOTTOMLEFT", 0, 64);
        self.sheetSelector:Show();
    end
    return self.frame;
end


function LibSpreadsheet:SetCellSpacing(size)
    if ( type(size) ~= "number" ) then error(("%s: Error in setting cell spacing; Expected nuber, got %s!"):format(SPREADSHEET, type(number))); end
    self.cellSpacing = number;
end

function LibSpreadsheet:SetCellPadding(size)
    if ( type(size) ~= "number" ) then error(("%s: Error in setting cell padding; Expected nuber, got %s!"):format(SPREADSHEET, type(number))); end
    self.cellPadding = number;
end

function LibSpreadsheet:SetInteractive(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting interactive mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
    self.isInteractive = toggle;
end

function LibSpreadsheet:ShowAxes(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting axis mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
    self.showAxes = toggle;
end

function LibSpreadsheet:SetColorRange(column, ...)
    column = self:ExtractLocation(column);
    if ( self.columns[column] == nil ) then error(("%s: Error while setting color range for column %s: No such column!"):format(SPREADSHEET, LibSpreadsheet.cols[column])); end
    self.columns[column].colorRange = {};
    for i = 1, select("#", ...) do
		local color = select(i, ...);
        table.insert(self.columns[column].colorRange, color);
    end
end


function LibSpreadsheet:Leaf(direction)
    local c = 1;
    for k, v in pairs(self.sheets) do
        if ( v.title == self.selectedSheet.title ) then
            c = k;
            break;
        end
    end
    if ( direction == 1 ) then
        c = c + 1;
        if ( self.sheets[c] ~= nil ) then
            self.selectedSheet = self.sheets[c];
        end
    elseif ( direction == -1 ) then
        c = c - 1;
        if ( self.sheets[c] ~= nil ) then
            self.selectedSheet = self.sheets[c];
        end
    end
    if ( c < 2 ) then
        self.prevSheet:Disable();
    else
        self.prevSheet:Enable();
    end
    if ( c == table.maxn(self.sheets) ) then
        self.nextSheet:Disable();
    else
        self.nextSheet:Enable();
    end
    self.sheetSelectorText:SetText(self.title .. ": " .. self.selectedSheet.title);
    self.verticalScrollbar:SetValue(0);
end

function LibSpreadsheet:SetVisibleSelection(fromCol, fromRow, toCol, toRow)
    if ( type(fromCol) == "number" ) then
        self.selectStartColumn = fromCol;
        self.selectStartRow = fromRow;
    elseif ( type(fromCol) == "string" ) then
        local c, r = self:ExtractLocation(fromCol);
        if ( type(r) == "number" ) then
            self.selectStartColumn = c;
            self.selectStartRow = r;
            toRow = toCol;
            toCol = fromRow;
        else
            self.selectStartColumn = c;
            self.selectStartRow = fromRow;
        end
    end
    if ( type(toCol) == "number" ) then
        self.selectEndColumn = toCol;
        self.selectEndRow = toRow;
    elseif ( type(toCol) == "string" ) then
        local c, r = self:ExtractLocation(toCol);
        if ( type(r) == "number" ) then
            self.selectEndColumn = c;
            self.selectEndRow = r;
        else
            self.selectEndColumn = c;
            self.selectEndRow = toRow;
        end
    end
    self.selecting = true;
    self:UpdateSelect(nil);
    self.selecting = false;
end


function LibSpreadsheet:StartSelect(caller, event)
    self.selecting = true;
    if ( IsShiftKeyDown() ) then
        self.selectEndColumn = caller.column;
        self.selectEndRow = caller.row;
    else
        self.selectStartColumn = caller.column;
        self.selectStartRow = caller.row;
    end
    if ( caller.row == true ) then
        self.selectStartRow = caller.index;
        self.selectStartColumn = 1;
        self.selectEndColumn = #self.columns;
    end
    if ( caller.column == true ) then
        self.selectStartColumn = caller.index;
        self.selectStartRow = 1;
        self.selectEndRow = #self.rows;
    end
    if ( event == "RightButton" ) then
        self.selectStartColumn = nil;
        self.selectStartRow = nil;
        self.selectEndColumn = nil;
        self.selectEndRow = nil;
    end
    self:UpdateSelect(caller);
end

function LibSpreadsheet:EndSelect(caller, event)
    self.selecting = false;
    if ( event == "LeftButton" ) then
        if not ( self.selectStartColumn == self.selectEndColumn and self.selectStartRow == self.selectEndRow ) then
            if ( self.callbacks['OnSelectionChange'] ~= nil ) then
                if ( type(self.callbacks['OnSelectionChange']) == "function" ) then
                    (self.callbacks['OnSelectionChange'])(math.min(self.selectStartColumn, self.selectEndColumn), math.min(self.selectStartRow, self.selectEndRow), math.max(self.selectStartColumn, self.selectEndColumn), math.max(self.selectStartRow, self.selectEndRow));
                end
            end
        end
    end
end

function LibSpreadsheet:UpdateSelect(caller)
    local rowTable = self.rows;
    if ( self.hasGroups or self.hasFilters ) then rowTable = self.rowHeap; end
    if ( self.selecting or false == true ) then
        if ( caller ~= nil ) then
            self.selectEndColumn = caller.column;
            self.selectEndRow = caller.row;
            if ( caller.row == true ) then
                self.selectEndColumn = #self.columns;
                self.selectEndRow = caller.index;
            end
            if ( caller.column == true ) then
                self.selectEndRow = #self.rows;
                self.selectEndColumn = caller.index;
            end
        end
        for x = 1, table.maxn(rowTable) do
            for y = 1, table.maxn(self.columns) do
                if ( rowTable[x].cells[y].frame ) then rowTable[x].cells[y].frame.overlay:Hide(); end
            end
        end
        if ( self.selectEndRow == nil ) then self.selectEndRow  = self.selectStartRow ; end
        if ( self.selectStartRow ~= nil ) then
            for x = math.min(self.selectStartRow, self.selectEndRow), math.max(self.selectStartRow, self.selectEndRow) do
                for y = math.min(self.selectStartColumn, self.selectEndColumn), math.max(self.selectStartColumn, self.selectEndColumn) do
                    if ( rowTable[x] and rowTable[x].cells[y] and rowTable[x].cells[y].frame ) then
                        rowTable[x].cells[y].frame.overlay:SetTexture(0.3,0.2,1,0.25);
                        rowTable[x].cells[y].frame.overlay:Show();
                    end
                end
            end
        end
    end
end

function LibSpreadsheet:HideColumn(...)
	for i = 1, select("#", ...) do
		local column = self:ExtractLocation(select(i, ...));
        if ( self.columns[column] ~= nil ) then
            self.columns[column].hidden = true;
        end
    end
end

function LibSpreadsheet:ShowColumn(...)
	for i = 1, select("#", ...) do
		local column = self:ExtractLocation(select(i, ...));
        if ( self.columns[column] ~= nil ) then
            self.columns[column].hidden = false;
        end
    end
end

function LibSpreadsheet:SetSheetCallback(event, func)
    local found = false;
    for k,v in pairs(LibSpreadsheet.callbackEvents) do
        if ( v == event ) then
            found = true;
            break;
        end
    end
    if ( found == true ) then
        if ( type(func) == "function" ) then
            self.callbacks[event] = func;
        else
            error(("%s: Error in SetCallback(): expected function, got %s!"):format(SPREADSHEET, type(func)));
        end
    else
        error(("%s: Error in SetCallback(): No such event, '%s'!"):format(SPREADSHEET, event));
    end
end

function LibSpreadsheet:ShowEditBox(cellFrame)
    local ret = true;
    if ( self.parent().editbox:IsVisible() ) then
        self.parent().editbox:Hide();
    end
    if ( self.callbacks['OnUserInputRequested'] ~= nil ) then
        if ( type(self.callbacks['OnUserInputRequested']) == "function" ) then
            ret = (self.callbacks['OnUserInputRequested'])(cellFrame.col, cellFrame.row, cellFrame.cell.id, cellFrame.cell.data);
        end
    end
    if ( ret ) then
        self.parent().editbox:SetText("");
        self.parent().editbox:ClearAllPoints();
        self.parent().editbox:SetTextInsets(0, 0, 3, 3);
        self.parent().editbox.caller = cellFrame;
        self.parent().editbox:SetAllPoints(cellFrame);
        self.parent().editbox:SetHeight(16);
        self.parent().editbox:SetFrameStrata("FULLSCREEN_DIALOG");
        self.parent().editbox:Show();
        cellFrame:Hide();
    end
end

function LibSpreadsheet:SaveEditBox(cellFrame, text)
    local ret = true;
    if ( self.callbacks['OnUserInputProcessed'] ~= nil ) then
        if ( type(self.callbacks['OnUserInputProcessed']) == "function" ) then
            ret = (self.callbacks['OnUserInputProcessed'])(cellFrame.col, cellFrame.row, cellFrame.cell.id, cellFrame.cell.data, text);
        end
    end
    if ( ret ) then
        cellFrame.cell.data = text;
        cellFrame.cell.dataChanged = true;
        self.parent():Render();
    end
    self.parent().editbox:ClearAllPoints();
    self.parent().editbox:SetText("");
end


--[[
LibSpreadsheet-1.0 Macro functions
]]--

function LibSpreadsheet:Macro_If(eval, iftrue, iffalse)
    if ( eval ) then
        return iftrue;
    else
        return iffalse;
    end
end

function LibSpreadsheet:Macro_And(expa, expb)
    return (expa and expb);
end

function LibSpreadsheet:Macro_Or(expa, expb)
    return (expa or expb);
end


function LibSpreadsheet:Macro_Sum(...)
    local val = 0;
    for k, v in pairs({select(1, ...)}) do
        val = val + (tonumber(v) or 0);
    end
    return val;
end

function LibSpreadsheet:Macro_Average(...)
    local val = 0;
    for k, v in pairs({select(1, ...)}) do
        val = val + (tonumber(v) or 0);
    end
    return val/select("#", ...);
end

function LibSpreadsheet:Macro_Mode(...)
    local vals = {select(1,...)};
    local mode = {};
    table.sort(vals);
    for k, v in pairs(vals) do
        mode[v] = (mode[v] or 0) + 1;
    end
    local n = 0;
    local m = 0;
    for k, v in pairs(mode) do
        if ( v > n ) then n = v; m = k; end
    end
    return m;
end


function LibSpreadsheet:Macro_Median(...)
    local vals = {select(1,...)};
    table.sort(vals);
    return vals[math.floor((#vals/2) + 0.5)] or 0;
end

function LibSpreadsheet:Macro_Large(k, ...)
    local vals = {select(1,...)};
    table.sort(vals);
    return vals[#vals-(k-1)] or 0;
end

function LibSpreadsheet:Macro_Small(k, ...)
    local vals = {select(1,...)};
    table.sort(vals);
    return vals[k] or 0;
end

function LibSpreadsheet:Macro_Sign(val)
    return ( val > 0 and 1 ) or ( val < 0 and -1 ) or 0;
end


function LibSpreadsheet:Macro_Logarithm(num, base)
    return (math.log(num) / math.log(base or 10));
end

function LibSpreadsheet:Macro_Unique(...)
    local tmp = {};
    local ret = {};
    for k, v in pairs({select(1, ...)}) do
        if ( tmp[v] == nil ) then
            tmp[v] = 1;
            table.insert(ret, v);
        end
    end
    return unpack(ret);
end

function LibSpreadsheet:Macro_StdDev(...)
    local val = 0;
    local population = select("#", ...);
    for k, v in pairs({select(1, ...)}) do
        val = val + (tonumber(v) or 0);
    end
    local mean = val / population;
    local val = 0;
    for k, v in pairs({select(1, ...)}) do
        val = val + math.pow( ( (tonumber(v) or 0) - mean), 2);
    end
    val = val /(population-1);
    return math.sqrt(val);
end

function LibSpreadsheet:Macro_Round(num, mul)
    return ( math.floor((num/(mul or 1))+0.5) * (mul or 1));
end

function LibSpreadsheet:Macro_Floor(num, mul)
    return ( math.floor((num/(mul or 1))) * (mul or 1));
end

function LibSpreadsheet:Macro_Ceil(num, mul)
    return ( math.ceil((num/(mul or 1))) * (mul or 1));
end


function LibSpreadTest()
    local lss = LibStub("LibSpreadsheet-1.0", true);                    -- Load the lib
    local book = lss:Book("LibSpreadsheet examples");                   -- Make a new book
    local sheet = book:Add("Food!");                                 -- Add a sheet to the book
    book:StrictMode(true);
    sheet.isInteractive = true;
    book:ShowSelector(true);
    sheet:AddColumn(
        {name="Name", sort=true},
        {name="Type", sort = true},
        {name="Is healthy", type="boolean"},
        {name="Calories", type="number"},
        {name="Some other column", type="string"}
        );  -- Set some headers
    local i;
    sheet:SetRow(sheet:AddRow({data="Apple", icon=[[Interface\LFGFrame\LFGRole]], texcoords={0.25,0.50,0,1}}, "Fruit", true,12), {background={0.3,0.8,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Peach", icon=[[Interface\Icons\Inv_misc_food_42]]}, "Fruit", true,12), {background={0.3,0.8,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Banana", icon=[[Interface\Icons\Inv_misc_food_24]]}, "Fruit", true,34), {background={0.3,0.8,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Orange", icon=[[Interface\Icons\Inv_misc_food_41]]}, "Fruit", true,47), {background={0.3,0.8,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Ham", icon=[[Interface\Icons\Inv_misc_food_13]]}, "Meat", true, 48), {background={0.8,0.4,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Salmon", icon=[[Interface\Icons\Inv_misc_food_75]]}, "Meat", true, 83), {background={0.5,0.6,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="T-bone Steak", icon=[[Interface\Icons\Inv_misc_food_70]]}, "Meat", false), {background={0.8,0.7,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Sausage", icon=[[Interface\Icons\Inv_misc_food_66]]}, "Meat", false), {background={0.8,0.3,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Croissant", icon=[[Interface\Icons\Inv_misc_food_32]]}, "Bread", false), {background={0.9,0.2,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Bagel", icon=[[Interface\Icons\Inv_misc_food_34]]}, "Bread", true), {background={0.7,0.6,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Wholegrain Bread", icon=[[Interface\Icons\Inv_misc_food_95_grainbread]]}, "Bread", true), {background={0.4,0.7,0.2,0.5}});
    sheet:SetRow(sheet:AddRow({data="Cinnamon Roll", icon=[[Interface\Icons\Inv_misc_food_73cinnamonroll]]}, "Bread", false, 100, "moooooooooo"), {background={0.9,0.3,0.2,0.5}});
    book:EnableMacros(true);
    book:ShowAxes(true);
    sheet = book:Select("Food!");
    if ( UIParent ) then
        local frame = book:Render();                                        -- Create the UI
        local window = CreateFrame("Frame", nil, UIParent, "DialogBoxFrame");
        window:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
        window:SetWidth(600);
        window:SetHeight(350);
        window:RegisterForDrag("LeftButton");
        window:SetScript("OnDragStart",function()  window:StartMoving(); end);
        window:SetScript("OnDragStop", function()  window:StopMovingOrSizing(); end);
        window:SetMovable(true);
        window:EnableMouse(true);
        frame:SetParent(window);
        frame:SetPoint("TOPLEFT", window, "TOPLEFT", 20, -20);
        frame:SetPoint("BOTTOMRIGHT", window, "BOTTOMRIGHT", -20, 50);
        frame:Show();
        window:Show();
    end
end

if ( dofile ) then
    LibSpreadTest();
end
